__all__ = ["ReassembleStream", "patch", "unpatch"]

import asyncio
import functools
from contextlib import contextmanager
from typing import Callable, List, Union

import openai
from openai import AsyncStream, Stream
from openai.types.chat import ChatCompletion
from packaging import version

from weave.monitoring.monitor import Monitor, Span, StatusCode, default_monitor
from weave.monitoring.openai.models import Context
from weave.monitoring.openai.util import Any, Context
from weave.wandb_interface.wandb_stream_table import StreamTable

from .models import *
from .util import *

from ..monitor import _get_global_monitor

old_create = openai.resources.chat.completions.Completions.create
old_async_create = openai.resources.chat.completions.AsyncCompletions.create

Callbacks = List[Callable]


class Callback:
    def before_send_request(self, context: Context, *args: Any, **kwargs: Any) -> None:
        ...

    def before_end(self, context: Context, *args: Any, **kwargs: Any) -> None:
        ...

    def before_yield_chunk(self, context: Context, *args: Any, **kwargs: Any) -> None:
        ...

    def after_yield_chunk(self, context: Context, *args: Any, **kwargs: Any) -> None:
        ...


class ReassembleStream(Callback):
    def before_send_request(self, context: Context, *args: Any, **kwargs: Any) -> None:
        sig = match_signature(old_create, *args, **kwargs)
        context.inputs = ChatCompletionRequest.parse_obj(sig)

    def before_end(self, context: Context, *args: Any, **kwargs: Any) -> None:
        if hasattr(context, "chunks") and context.inputs is not None:
            input_messages = context.inputs.messages
            context.outputs = reconstruct_completion(input_messages, context.chunks)  # type: ignore


class AsyncChatCompletions:
    def __init__(
        self,
        base_create: Callable,
        callbacks: Optional[List[Callback]] = None,
        streamtable: Optional[StreamTable] = None,
    ) -> None:
        self._base_create = base_create
        if callbacks is None:
            callbacks = make_default_callbacks()
        self.callbacks = callbacks
        self.monitor = Monitor(streamtable)

    async def create(
        self, *args: Any, **kwargs: Any
    ) -> Union[ChatCompletion, AsyncStream[ChatCompletionChunk]]:
        self.context = Context()
        if kwargs.get("stream", False):
            return self._streaming_create(*args, **kwargs)
        return await self._create(*args, **kwargs)

    async def _create(self, *args: Any, **kwargs: Any) -> ChatCompletion:
        with span_context(self.monitor, self.context, "request", *args, **kwargs):
            await self._use_callbacks("before_send_request", *args, **kwargs)

            result = await self._base_create(*args, **kwargs)
            self.context.outputs = result

            await self._use_callbacks("before_end", *args, **kwargs)

            return result

    async def _streaming_create(
        self, *args: Any, **kwargs: Any
    ) -> AsyncStream[ChatCompletionChunk]:
        with span_context(self.monitor, self.context, "request", *args, **kwargs):
            await self._use_callbacks("before_send_request", *args, **kwargs)
            for callback in self.callbacks:
                await self._use_callback(
                    callback.before_send_request, self.context, *args, **kwargs
                )

            stream = await self._base_create(*args, **kwargs)
            self.context.chunks = []  # type: ignore
            async for chunk in stream:
                await self._use_callbacks("before_yield_chunk", *args, **kwargs)
                yield chunk
                self.context.chunks.append(chunk)  # type: ignore
                await self._use_callbacks("after_yield_chunk", *args, **kwargs)
            await self._use_callbacks("before_end", *args, **kwargs)

    @staticmethod
    async def _use_callback(
        f: Callable, context: Context, *args: Any, **kwargs: Any
    ) -> None:
        if asyncio.iscoroutinefunction(f):
            await f(context, *args, **kwargs)
        else:
            f(context, *args, **kwargs)

    async def _use_callbacks(self, step: str, *args: Any, **kwargs: Any) -> None:
        for callback in self.callbacks:
            try:
                method = getattr(callback, step)
            except AttributeError:
                error("Invalid callback.  Did you forget to inherit from Callback?")
                return

            await self._use_callback(method, self.context, *args, **kwargs)


class ChatCompletions:
    def __init__(
        self,
        base_create: Callable,
        callbacks: Optional[List[Callback]] = None,
        streamtable: Optional[StreamTable] = None,
    ) -> None:
        self._base_create = base_create
        if callbacks is None:
            callbacks = make_default_callbacks()
        self.callbacks = callbacks
        self.monitor = Monitor(streamtable)

    def create(
        self, *args: Any, **kwargs: Any
    ) -> Union[ChatCompletion, Stream[ChatCompletionChunk]]:
        self.context = Context()
        if kwargs.get("stream", False):
            result = self._streaming_create(*args, **kwargs)
            return result
        return self._create(*args, **kwargs)

    def _create(self, *args: Any, **kwargs: Any) -> ChatCompletion:
        with span_context(self.monitor, self.context, "request", *args, **kwargs):
            self._use_callbacks("before_send_request", *args, **kwargs)

            result = self._base_create(*args, **kwargs)
            self.context.outputs = result

            self._use_callbacks("before_end", *args, **kwargs)

            return result

    def _streaming_create(
        self, *args: Any, **kwargs: Any
    ) -> Stream[ChatCompletionChunk]:
        with span_context(self.monitor, self.context, "request", *args, **kwargs):
            self._use_callbacks("before_send_request", *args, **kwargs)

            stream = self._base_create(*args, **kwargs)
            self.context.chunks = []  # type: ignore
            for chunk in stream:
                self._use_callbacks("before_yield_chunk", *args, **kwargs)
                yield chunk
                self.context.chunks.append(chunk)  # type: ignore
                self._use_callbacks("after_yield_chunk", *args, **kwargs)

            self._use_callbacks("before_end", *args, **kwargs)

    def _use_callbacks(self, step: str, *args: Any, **kwargs: Any) -> None:
        for callback in self.callbacks:
            try:
                method = getattr(callback, step)
            except AttributeError:
                error("Invalid callback.  Did you forget to inherit from Callback?")
                return

            try:
                method(self.context, *args, **kwargs)
            except Exception as exception:
                warn(f"problem with {callback=}, {exception=}")


def patch(
    *,
    callbacks: Optional[List[Callback]] = None,
) -> None:
    def _patch() -> None:
        unpatch_fqn = f"{unpatch.__module__}.{unpatch.__qualname__}()"

        if _get_global_monitor() is not None:
            info(f"Patching OpenAI completions.  To unpatch, call {unpatch_fqn}")

            mon = default_monitor()

            hooks = ChatCompletions(
                old_create, callbacks=callbacks, streamtable=mon.streamtable
            )
            async_hooks = AsyncChatCompletions(
                old_async_create, callbacks=callbacks, streamtable=mon.streamtable
            )
            openai.resources.chat.completions.Completions.create = (
                functools.partialmethod(hooks.create)
            )
            openai.resources.chat.completions.AsyncCompletions.create = (
                functools.partialmethod(async_hooks.create)
            )

    if version.parse(openai.__version__) < version.parse("1.0.0"):
        error(
            f"this integration requires openai>=1.0.0 (got {openai.__version__}).  Please upgrade and try again"
        )
        return

    try:
        _patch()
    except Exception as e:
        error(f"problem patching: {e}, auto-unpatching")
        unpatch()
        raise Exception from e


def unpatch() -> None:
    if _get_global_monitor() is not None:
        info("Unpatching OpenAI completions")
        openai.resources.chat.completions.Completions.create = old_create
        openai.resources.chat.completions.AsyncCompletions.create = old_async_create


def make_default_callbacks() -> List[Callback]:
    return [ReassembleStream()]


@contextmanager
def span_context(
    monitor: Monitor, context: Context, span_name: str, *args: Any, **kwargs: Any
) -> Iterator[Span]:
    with monitor.span(span_name) as span:
        context.span = span
        try:
            yield span
        except Exception as e:
            span.status_code = StatusCode.ERROR
            span.exception = e
        finally:
            sig = match_signature(old_create, *args, **kwargs)
            span.inputs = ChatCompletionRequest.parse_obj(sig).dict()
            if context.outputs is not None:
                span.output = context.outputs.dict()



import typing

import weave

from .dataset import Dataset
from .structured_output import StructuredOutputChatModel


@weave.type()
class Evaluate:
    @weave.op()
    def compute(self, dataset: Dataset, predictions: list[typing.Any]) -> typing.Any:
        ...


@weave.type()
class EvaluateExactMatch(Evaluate):
    format_example: typing.Callable[[typing.Any], typing.Any]

    @weave.op()
    def compute(self, dataset: Dataset, predictions: list[typing.Any]) -> typing.Any:
        match = [p == self.format_example(r) for p, r in zip(predictions, dataset.rows)]
        return {
            "columns": {"match": match},
            "summary": {"match_frac": sum(match) / len(match)},
        }


@weave.type()
class EvaluateLLM(Evaluate):
    chat_llm: StructuredOutputChatModel
    messages_template: typing.Callable[[typing.Any, typing.Any], typing.Any]

    @weave.op()
    def compute(self, dataset: Dataset, predictions: list[typing.Any]) -> typing.Any:
        result_type = weave.types.TypedDict(
            {
                "score": weave.types.Float(),
                "rationale": weave.types.String(),
            }
        )
        from .. import weave_internal

        # TODO: get rid of this const() wrap... why do I need it?
        all_messages = dataset.rows.apply(  # type: ignore
            lambda r, i: self.messages_template(
                r, weave_internal.const(predictions)[i]  # type: ignore
            )
        )
        results = all_messages.apply(lambda m: self.chat_llm.complete(m, result_type))

        scores = results.column("score")
        rationales = results.column("rationale")

        non_none_scores = [s for s in scores if s is not None]
        score_avg = 0
        if len(non_none_scores) > 0:
            score_avg = sum(non_none_scores) / len(non_none_scores)
        return {
            "columns": {"score": scores, "rationale": rationales},
            "summary": {"score_avg": score_avg},
        }


@weave.type()
class EvaluateMulti(Evaluate):
    evaluations: dict[str, Evaluate]

    @weave.op()
    def compute(self, dataset: Dataset, predictions: list[typing.Any]) -> typing.Any:
        columns = {}
        summary = {}
        for k, eval in self.evaluations.items():
            eval_result = eval.compute(dataset, predictions)
            columns[k] = eval_result["columns"]
            summary[k] = eval_result["summary"]
        return {
            "columns": columns,
            "summary": summary,
        }


@weave.type()
class EvaluateNamedSteps(Evaluate):
    step_keys: list[str]
    evaluation: Evaluate

    @weave.op()
    def compute(self, dataset: typing.Any, predictions: list[typing.Any]) -> typing.Any:
        # This handles nested columns by recursively computing column deltas.
        # TODO: This would be nicer with column traversal / leaf-mapping.
        def compute_delta(
            cur: dict, prior: dict, delta_columns: dict, delta_summary: dict
        ) -> None:
            for k, v in cur.items():
                if isinstance(v, dict):
                    delta_columns[k] = {}
                    delta_summary[k] = {}
                    compute_delta(
                        v, prior.get(k, {}), delta_columns[k], delta_summary[k]
                    )
                else:
                    try:
                        float_delta = [
                            float(v[i]) - float(prior[k][i]) for i in range(len(v))
                        ]
                        delta_columns[k] = float_delta
                        delta_summary[k] = sum(float_delta) / len(float_delta)
                    except (ValueError, TypeError, KeyError):
                        continue

        columns = {}
        summary = {}
        for i, step_key in enumerate(self.step_keys):
            eval_result = self.evaluation.compute(
                dataset, [p[step_key] for p in predictions]
            )
            columns[step_key] = eval_result["columns"]
            summary[step_key] = eval_result["summary"]
            if i != 0:
                prior_columns = columns[self.step_keys[i - 1]]
                cur_columns = columns[step_key]
                delta_columns: dict = {}
                delta_summary: dict = {}
                compute_delta(cur_columns, prior_columns, delta_columns, delta_summary)
                columns[f"{step_key}_delta"] = delta_columns
                summary[f"{step_key}_delta_avg"] = delta_summary

        return {
            "columns": columns,
            "summary": summary,
        }



from weave import context_state

loading_builtins_token = context_state.set_loading_built_ins()

try:
    from .rep import *
finally:
    context_state.clear_loading_built_ins(loading_builtins_token)



from pydantic import BaseModel, create_model
from . import weave_types as types


def weave_type_to_pydantic(
    property_types: dict[str, types.Type], name: str
) -> BaseModel:
    field_types = {}
    for k, v in property_types.items():
        if isinstance(v, types.TypedDict):
            field_types[k] = weave_type_to_pydantic(v.property_types, "TypedDict")
        else:
            instance_classes = v._instance_classes()
            if not instance_classes:
                raise ValueError(f"Cannot convert {v} to pydantic")
            field_types[k] = instance_classes[0]  # type: ignore
    field_defs = {k: (v, ...) for k, v in field_types.items()}
    return create_model(name, **field_defs)  # type: ignore



import weave

from .. import weave_types as types
from ..panels import panel_board
from .. import weave_internal
from .generator_templates import template_registry

from ..ops_primitives import list_


panels = weave.panels


BOARD_INPUT_WEAVE_TYPE = types.List(
    types.TypedDict(
        {
            "timestamp": types.optional(types.Timestamp()),
            "_timestamp": types.Number(),
            "entity_name": types.optional(types.String()),
            "project_name": types.optional(types.String()),
            "queue_uri": types.optional(types.String()),
            "sweep_id": types.optional(types.String()),
            "run_id": types.optional(types.String()),
            "job": types.optional(types.String()),
            "trace_id": types.optional(types.String()),
            "state": types.optional(types.String()),
            "error": types.optional(types.String()),
            "metrics": types.TypedDict(
                {
                    "system": types.TypedDict(
                        {
                            "tpu": types.optional(types.Number()),
                            "cpu": types.optional(types.Number()),
                            "cpu_cores_util": types.optional(
                                types.List(types.Number())
                            ),
                            "gpu_cores_util": types.optional(
                                types.List(types.Number())
                            ),
                            "gpu_cores_mem": types.optional(types.List(types.Number())),
                            "memory": types.optional(types.Number()),
                            "disk": types.TypedDict(
                                {
                                    "in": types.optional(types.Number()),
                                    "out": types.optional(types.Number()),
                                }
                            ),
                            "proc": types.TypedDict(
                                {
                                    "memory": types.TypedDict(
                                        {
                                            "availableMB": types.optional(
                                                types.Number()
                                            ),
                                        }
                                    )
                                }
                            ),
                        }
                    )
                }
            ),
        }
    )
)


@weave.op(  # type: ignore
    name="py_board-observability",
    hidden=False,
    input_type={
        "input_node": types.Function(
            {},
            BOARD_INPUT_WEAVE_TYPE,
        )
    },
)
def observability(
    input_node,
) -> panels.Group:
    timestamp_col_name = "timestamp"
    num_buckets = 80

    now = weave.ops.datetime_now()

    dashboard = weave.panels.Group(
        layoutMode="grid",
        showExpressions=False,
        enableAddPanel=False,
        disableDeletePanel=True,
    )

    varbar = panel_board.varbar(editable=False)
    source_data = input_node

    filter_fn = varbar.add(
        "filter_fn",
        weave_internal.define_fn(
            {"row": source_data.type.object_type},
            lambda row: weave_internal.const(True),
        ),
        hidden=True,
    )

    grouping_fn = varbar.add(
        "grouping_fn",
        weave_internal.define_fn(
            {"row": input_node.type.object_type}, lambda row: row["entity_name"]
        ),
        hidden=True,
    )

    three_days_in_seconds = 60 * 60 * 24 * 3
    window_start = varbar.add(
        "window_start",
        weave.ops.from_number(now - three_days_in_seconds),
        hidden=True,
    )
    window_end = varbar.add(
        "window_end",
        weave.ops.from_number(now),
        hidden=True,
    )

    filtered_range = varbar.add(
        "filtered_range",
        weave.ops.make_list(
            a=window_start,
            b=window_end,
        ),
        hidden=True,
    )

    user_zoom_range = varbar.add("user_zoom_range", None, hidden=True)
    varbar.add(
        "Time_range",
        panels.DateRange(user_zoom_range, domain=filtered_range),
    )
    bin_range = varbar.add(
        "bin_range", user_zoom_range.coalesce(filtered_range), hidden=True
    )

    window_data = source_data.filter(
        lambda row: weave.ops.Boolean.bool_and(
            row[timestamp_col_name] >= bin_range[0],
            row[timestamp_col_name] <= bin_range[1],
        )
    )
    varbar.add(
        "Grouping",
        panels.GroupingEditor(grouping_fn, node=window_data),
    )
    filtered_window_data = varbar.add(
        "filtered_window_data",
        window_data.filter(filter_fn),
        hidden=True,
    )

    varbar.add(
        "Filters",
        panels.FilterEditor(filter_fn, node=window_data),
    )

    colors_node = weave.ops.dict_(
        **{
            "running": "rgb(57, 126, 237)",
            "failed": "rgb(255, 122, 136)",
            "crashed": "rgb(255, 61, 90)",
            "finished": "rgb(0, 178, 110)",
            "starting": "rgb(125, 177, 250)",
            "failed_rqi": "rgb(255, 199, 202)",
            "queued": "rgb(211, 215, 222)",
            "popped": "rgb(189, 217, 255)",
        }
    )

    state_color_func = weave_internal.define_fn(
        {"row": input_node.type.object_type},
        lambda row: colors_node.pick(row["state"]),
    )

    queued_time_data = varbar.add(
        "queued_time_data",
        filtered_window_data.filter(
            lambda row: weave.ops.Boolean.bool_or(
                row["state"] == "queued",
                row["state"] == "starting",
            ),
        ),
        hidden=True,
    )

    is_start_stop_state = weave_internal.define_fn(
        {"row": input_node.type.object_type},
        lambda row: weave.ops.Boolean.bool_or(
            row["state"] == "running",
            weave.ops.Boolean.bool_or(
                row["state"] == "finished",
                weave.ops.Boolean.bool_or(
                    row["state"] == "crashed",
                    row["state"] == "failed",
                ),
            ),
        ),
    )

    state_transitions_plot = panels.Plot(
        filtered_window_data,
        x=lambda row: row[timestamp_col_name].bin(
            weave.ops.timestamp_bins_nice(bin_range, num_buckets)
        ),
        x_title="Time",
        y=lambda row: row.count(),
        y_title="Count of transitions by state",
        label=lambda row: row["state"],
        tooltip=lambda row: weave.ops.dict_(
            **{
                "state": row["state"][0],
                "count": row.count(),
            }
        ),
        color_title="state",
        color=state_color_func.val,
        groupby_dims=["x", "label"],
        mark="bar",
        domain_x=user_zoom_range,
    )

    queued_time_plot = panels.Plot(
        queued_time_data,
        x=lambda row: row[timestamp_col_name].bin(
            weave.ops.timestamp_bins_nice(bin_range, num_buckets / 2)
        ),
        x_title="Time",
        y=lambda row: weave.ops.timedelta_total_seconds(
            weave.ops.datetime_sub(
                row[timestamp_col_name].max(), row[timestamp_col_name].min()
            )
        ),
        y_title="Time spent queued",
        label=lambda row: grouping_fn(row),
        tooltip=lambda row: weave.ops.dict_(
            **{
                "job": row["job"][0],
                "user": row["entity_name"][0],
                "project": row["project_name"][0],
                "duration (s)": weave.ops.timedelta_total_seconds(
                    weave.ops.datetime_sub(
                        row[timestamp_col_name].max(), row[timestamp_col_name].min()
                    )
                ),
            }
        ),
        color_title="Grouping",
        groupby_dims=["x", "label"],
        mark="bar",
        domain_x=user_zoom_range,
    )

    start_stop_states = varbar.add(
        "start_stop_states",
        weave.ops.List.filter(filtered_window_data, is_start_stop_state),
        hidden=True,
    )

    latest_runs = varbar.add(
        "latest_runs",
        weave.ops.List.limit(
            weave.ops.List.sort(
                arr=start_stop_states,
                compFn=lambda row: weave.ops.make_list(a=row["timestamp"]),
                columnDirs=["desc"],
            ),
            40,
        ),
        hidden=True,
    )

    latest_runs_plot = panels.Plot(
        latest_runs,
        x=lambda row: row[timestamp_col_name],
        x_title="Time",
        y_title="Run ID",
        y=lambda row: row["run_id"],
        tooltip=lambda row: row[0]["job"],
        label=lambda row: row["run_id"],
        groupby_dims=["x", "y"],
        mark="line",
        no_legend=True,
        domain_x=user_zoom_range,
    )

    jobs = varbar.add(
        "jobs",
        filtered_window_data.filter(is_start_stop_state),
        hidden=True,
    )
    jobs_table = panels.Table(jobs)  # type: ignore
    jobs_table.add_column(lambda row: row["run_id"], "Run ID", groupby=True)
    jobs_table.add_column(
        lambda row: weave.ops.timedelta_total_seconds(
            weave.ops.datetime_sub(
                row[timestamp_col_name].max(), row[timestamp_col_name].min()
            )
        ),
        "runtime (s)",
        sort_dir="desc",
    )
    jobs_table.add_column(lambda row: row["job"][0], "Job")  # groupby=True
    jobs_table.add_column(
        lambda row: row["metrics"]["system"]["cpu_cores_util"][-1].avg(),
        "avg cpu util %",
    )

    starting_runs = varbar.add(
        "starting_runs",
        filtered_window_data.filter(
            weave_internal.define_fn(
                {"row": source_data.type.object_type},
                lambda row: row["state"] == "starting",
            )
        ),
        hidden=True,
    )
    runs_table = panels.Table(starting_runs)  # type: ignore
    runs_table.add_column(lambda row: row["entity_name"], "User", groupby=True)
    runs_table.add_column(lambda row: row.count(), "Count", sort_dir="desc")

    finished_runs = varbar.add(
        "finished_runs",
        filtered_window_data.filter(
            lambda row: weave.ops.Boolean.bool_or(
                row["state"] == "finished",
                weave.ops.Boolean.bool_or(
                    row["state"] == "crashed",
                    row["state"] == "failed",
                ),
            ),
        ),
        hidden=True,
    )

    runs_by_user_project_plot = panels.Plot(
        finished_runs,
        x=lambda row: row["project_name"],
        x_title="Grouping",
        y_title="Count by user",
        y=lambda row: row.count(),
        label=lambda row: grouping_fn(row),
        groupby_dims=["x", "label"],
        mark="bar",
        no_legend=True,
    )

    gpu_waste_by_user_plot = panels.Plot(
        start_stop_states,
        x=lambda row: weave.ops.Number.__mul__(
            weave.ops.timedelta_total_seconds(
                weave.ops.datetime_sub(
                    row[timestamp_col_name].max(), row[timestamp_col_name].min()
                ),
            ),
            1000,
        ),
        x_title="Run duration",
        y_title="Gpu usage (%)",
        y=lambda row: row["metrics"]["system"]["gpu_cores_util"][-1].avg(),
        tooltip=lambda row: weave.ops.dict_(
            **{
                "Run": weave.ops.join_to_str(
                    weave.ops.make_list(
                        a=row["entity_name"][0],
                        b=row["project_name"][0],
                        c=row["run_id"][0],
                    ),
                    "/",
                ),
                "Job": row["job"][0],
                "Duration (s) * (1 - gpu %)": weave.ops.Number.__mul__(
                    weave.ops.Number.__mul__(
                        weave.ops.timedelta_total_seconds(
                            weave.ops.datetime_sub(
                                row[timestamp_col_name].max(),
                                row[timestamp_col_name].min(),
                            ),
                        ),
                        1000,
                    ),
                    weave.ops.Number.__sub__(
                        1,
                        weave.ops.Number.__truediv__(
                            row["metrics"]["system"]["gpu_cores_util"][-1].avg(),
                            100,
                        ),
                    ),
                ),
                "Duration (s)": weave.ops.Number.__mul__(
                    weave.ops.timedelta_total_seconds(
                        weave.ops.datetime_sub(
                            row[timestamp_col_name].max(), row[timestamp_col_name].min()
                        ),
                    ),
                    1000,
                ),
                "Gpu util %": row["metrics"]["system"]["gpu_cores_util"][-1].avg(),
            }
        ),
        label=lambda row: row["run_id"],
        groupby_dims=["label"],
        mark="point",
        no_legend=True,
        domain_y=weave_internal.make_const_node(types.List(types.Number()), [0, 100]),
    )

    def make_metric_plot(metric_name: str, y_title: str) -> panels.Plot:
        return panels.Plot(
            finished_runs,
            x=lambda row: row[timestamp_col_name],
            x_title=timestamp_col_name,
            y=lambda row: list_.List.concat(
                row["metrics"]["system"][metric_name]
            ).avg(),
            y_title=y_title,
            tooltip=lambda row: weave.ops.dict_(
                **{
                    "user": row["entity_name"][0],
                    "run id": row["run_id"][-1],
                    "job": row["job"][0],
                    "project": row["project_name"][0],
                    "min": row["metrics"]["system"][metric_name][0].min(),
                    "max": row["metrics"]["system"][metric_name][0].max(),
                    "avg": row["metrics"]["system"][metric_name][0].avg(),
                }
            ),
            groupby_dims=["x"],
            mark="line",
            no_legend=True,
            domain_x=bin_range,
            domain_y=weave_internal.make_const_node(
                types.List(types.Number()), [0, 100]
            ),
        )

    cpu_plot = make_metric_plot("cpu_cores_util", "avg cpu %")
    gpu_plot = make_metric_plot("gpu_cores_util", "avg gpu %")
    gpu_memory_plot = make_metric_plot("gpu_cores_mem", "avg gpu memory util %")
    memory_plot = weave.panels.Plot(
        finished_runs,
        x=lambda row: row[timestamp_col_name],
        x_title=timestamp_col_name,
        y=lambda row: row["metrics"]["system"]["memory"][0],
        y_title="system memory used (MB)",
        tooltip=lambda row: weave.ops.dict_(
            **{
                "user": row["entity_name"][0],
                "run id": row["run_id"][-1],
                "job": row["job"][0],
                "project": row["project_name"][0],
                "memory": row["metrics"]["system"]["memory"][0],
            }
        ),
        groupby_dims=["x"],
        mark="line",
        no_legend=True,
        domain_x=user_zoom_range,
    )

    errors_table = panels.Table(  # type: ignore
        filtered_window_data.filter(
            weave_internal.define_fn(
                {"row": source_data.type.object_type},
                lambda row: row["error"] != None,
            )
        )
    )
    errors_table.add_column(lambda row: row["run_id"], "Run ID")
    errors_table.add_column(lambda row: row["timestamp"], "Timestamp", sort_dir="desc")
    errors_table.add_column(lambda row: row["job"], "Job")
    errors_table.add_column(lambda row: row["error"], "Error", panel_def="object")

    # layout
    dashboard.add(
        "Job_status",
        state_transitions_plot,
        layout=panels.GroupPanelLayout(x=0, y=0, w=24, h=6),
    )
    dashboard.add(
        "Queued_time",
        queued_time_plot,
        layout=panels.GroupPanelLayout(x=0, y=6, w=12, h=8),
    )
    dashboard.add(
        "Lastest_runs",
        latest_runs_plot,
        layout=panels.GroupPanelLayout(x=12, y=6, w=12, h=8),
    )
    dashboard.add(
        "Longest_jobs",
        jobs_table,
        layout=panels.GroupPanelLayout(x=0, y=14, w=12, h=8),
    )
    dashboard.add(
        "Runs_by_user",
        runs_table,
        layout=panels.GroupPanelLayout(x=12, y=14, w=12, h=8),
    )
    dashboard.add(
        "Gpu_waste_by_user",
        gpu_waste_by_user_plot,
        layout=panels.GroupPanelLayout(x=0, y=22, w=12, h=6),
    )
    dashboard.add(
        "Runs_by_project",
        runs_by_user_project_plot,
        layout=panels.GroupPanelLayout(x=12, y=22, w=12, h=6),
    )
    dashboard.add(
        "Cpu_usage_on_run_finish",
        cpu_plot,
        layout=panels.GroupPanelLayout(x=0, y=28, w=12, h=6),
    )
    dashboard.add(
        "System_memory_on_run_finish",
        memory_plot,
        layout=panels.GroupPanelLayout(x=12, y=28, w=12, h=6),
    )
    dashboard.add(
        "Gpu_usage_on_run_finish",
        gpu_plot,
        layout=panels.GroupPanelLayout(x=0, y=34, w=12, h=6),
    )
    dashboard.add(
        "Gpu_memory_on_run_finish",
        gpu_memory_plot,
        layout=panels.GroupPanelLayout(x=12, y=34, w=12, h=6),
    )
    dashboard.add(
        "Errors",
        errors_table,
        layout=panels.GroupPanelLayout(x=0, y=40, w=24, h=8),
    )

    return panels.Board(vars=varbar, panels=dashboard, editable=False)


template_registry.register(
    "py_board-observability",
    "Launch Observability",
    "Seed a board with an observability view to track Launch queue performance.",
)



import time
import typing
import os
import dataclasses

from .urls import BROWSE3_PATH
from . import graph as _graph
from . import graph_mapper as _graph_mapper
from . import storage as _storage
from . import ref_base as _ref_base
from . import artifact_wandb as _artifact_wandb
from . import wandb_api as _wandb_api
from . import trace as _trace
from . import weave_internal as _weave_internal
from . import errors as _errors
from . import ops as _ops
from . import util as _util
from . import context as _context
from . import context_state as _context_state
from . import run as _run
from . import graph_client as _graph_client
from . import graph_client_local as _graph_client_local
from . import graph_client_wandb_art_st as _graph_client_wandb_art_st
from . import graph_client_context as _graph_client_context
from weave import monitoring as _monitoring
from weave.monitoring import monitor as _monitor

# exposed as part of api
from . import weave_types as types
from . import types_numpy as _types_numpy
from . import errors
from .decorators import weave_class, op, mutation, type
from .op_args import OpVarArgs
from .op_def import OpDef
from . import usage_analytics
from .context import (
    use_fixed_server_port,
    use_frontend_devmode,
    # eager_execution,
    # lazy_execution,
)
from .server import capture_weave_server_logs
from .val_const import const
from .file_base import File, Dir
from .dispatch import RuntimeConstNode

from .weave_internal import define_fn

Node = _graph.Node


def save(node_or_obj, name=None):
    if isinstance(node_or_obj, _graph.Node):
        return _ops.save(node_or_obj, name=name)
    else:
        # If the user does not provide a branch, then we explicitly set it to
        # the default branch, "latest".
        branch = None
        name_contains_branch = name is not None and ":" in name
        if not name_contains_branch:
            branch = "latest"
        ref = _storage.save(node_or_obj, name=name, branch=branch)
        if name is None:
            # if the user didn't provide a name, the returned reference
            # will be to the specific version
            uri = ref.uri
        else:
            # otherwise the reference will be to whatever branch was provided
            # or the "latest" branch if only a name was provided.
            uri = ref.branch_uri
        return _ops.get(str(uri))


def get(ref_str):
    obj = _storage.get(ref_str)
    ref = typing.cast(_ref_base.Ref, _storage._get_ref(obj))
    return _weave_internal.make_const_node(ref.type, obj)


def use(nodes, client=None):
    usage_analytics.use_called()
    if client is None:
        client = _context.get_client()
    return _weave_internal.use(nodes, client)


def _get_ref(obj):
    if isinstance(obj, _storage.Ref):
        return obj
    ref = _storage.get_ref(obj)
    if ref is None:
        raise _errors.WeaveApiError("obj is not a weave object: %s" % obj)
    return ref


def versions(obj):
    if isinstance(obj, _graph.ConstNode):
        obj = obj.val
    elif isinstance(obj, _graph.OutputNode):
        obj = use(obj)
    ref = _get_ref(obj)
    return ref.versions()  # type: ignore


def expr(obj):
    ref = _get_ref(obj)
    return _trace.get_obj_expr(ref)


def type_of(obj: typing.Any) -> types.Type:
    return types.TypeRegistry.type_of(obj)


def weave(obj: typing.Any) -> RuntimeConstNode:
    return _weave_internal.make_const_node(type_of(obj), obj)  # type: ignore


def from_pandas(df):
    return _ops.pandas_to_awl(df)


#### Newer API below


def init(project_name: str) -> _graph_client.GraphClient:
    from . import wandb_api

    fields = project_name.split("/")
    if len(fields) == 1:
        api = wandb_api.get_wandb_api_sync()
        try:
            entity_name = api.default_entity_name()
        except AttributeError:
            raise errors.WeaveWandbAuthenticationException(
                'weave init requires wandb. Run "wandb login"'
            )
        project_name = fields[0]
    elif len(fields) == 2:
        entity_name, project_name = fields
    else:
        raise ValueError(
            'project_name must be of the form "<project_name>" or "<entity_name>/<project_name>"'
        )
    if not entity_name:
        raise ValueError("entity_name must be non-empty")
    client = _graph_client_wandb_art_st.GraphClientWandbArtStreamTable(
        entity_name, project_name
    )
    _graph_client_context._graph_client.set(client)
    print("Ensure you have the prototype UI running with `weave ui`")
    print(
        f"View project at http://localhost:3000/{BROWSE3_PATH}/{entity_name}/{project_name}"
    )
    return client


# This is currently an internal interface. We'll expose something like it though ("offline" mode)
def init_local_client() -> _graph_client.GraphClient:
    client = _graph_client_local.GraphClientLocal()
    _graph_client_context._graph_client.set(client)
    return client


def publish(obj: typing.Any, name: str) -> _ref_base.Ref:
    client = _graph_client_context.require_graph_client()

    ref = client.save_object(obj, name, "latest")

    print(f"Published {ref.type.root_type_class().name} to {ref.ui_url}")

    # Have to manually put the ref on the obj, this is supposed to happen at
    # a lower level, but we use a mapper in _direct_publish that I think changes
    # the object identity
    _ref_base._put_ref(obj, ref)
    return ref


def ref(uri: str) -> _ref_base.Ref:
    if not "://" in uri:
        client = _graph_client_context.get_graph_client()
        if not client:
            raise ValueError("Call weave.init() first, or pass a fully qualified uri")
        if "/" in uri:
            raise ValueError("'/' not currently supported in short-form URI")
        if ":" not in uri:
            name = uri
            version = "latest"
        else:
            name, version = uri.split(":")
        uri = str(client.ref_uri(name, version))

    return _ref_base.Ref.from_str(uri)


import contextlib


@contextlib.contextmanager
def attributes(attributes: typing.Dict[str, typing.Any]) -> typing.Iterator:
    cur_attributes = {**_monitor._attributes.get()}
    cur_attributes.update(attributes)

    token = _monitor._attributes.set(cur_attributes)
    try:
        yield
    finally:
        _monitor._attributes.reset(token)


def serve(
    model_ref: _artifact_wandb.WandbArtifactRef,
    method_name: typing.Optional[str] = None,
    auth_entity: typing.Optional[str] = None,
    port: int = 9996,
    thread: bool = False,
) -> typing.Optional[str]:
    import uvicorn
    from .serve_fastapi import object_method_app

    client = _graph_client_context.require_graph_client()
    if not isinstance(
        client, _graph_client_wandb_art_st.GraphClientWandbArtStreamTable
    ):
        raise ValueError("serve currently only supports wandb client")

    print(f"Serving {model_ref}")
    print(f"Server docs at http://localhost:{port}/docs")
    os.environ["PROJECT_NAME"] = f"{client.entity_name}/{client.project_name}"
    os.environ["MODEL_REF"] = str(model_ref)

    wandb_api_ctx = _wandb_api.get_wandb_api_context()
    app = object_method_app(model_ref, method_name=method_name, auth_entity=auth_entity)
    trace_attrs = _monitor._attributes.get()

    def run():
        with _wandb_api.wandb_api_context(wandb_api_ctx):
            with _context_state.eager_execution():
                with _context.execution_client():
                    with attributes(trace_attrs):
                        uvicorn.run(app, host="0.0.0.0", port=port)

    if _util.is_notebook():
        thread = True
    if thread:
        from threading import Thread

        t = Thread(target=run, daemon=True)
        t.start()
        time.sleep(1)
        return "http://localhost:%d" % port
    else:
        run()
    return None



from . import panel_llm_monitor
from . import panel_trace_monitor
from . import panel_autoboard
from . import panel_seedboard
from . import panel_observability
from . import generator_templates

# This is just an example, uncomment to enable
# from . import panel_example_custom_board



import inspect
import typing


from . import registry_mem
from . import op_def
from . import op_args
from . import derive_op
from . import context_state
from . import weave_types as types
from . import pyfunc_type_util

if typing.TYPE_CHECKING:
    from weave.gql_op_plugin import GqlOpPlugin


def op(
    input_type: pyfunc_type_util.MaybeInputTypeType = None,
    # input_type: pyfunc_type_util.MaybeInputTypeType = None,
    output_type: pyfunc_type_util.MaybeOutputTypeType = None,
    refine_output_type: typing.Optional[op_def.OpDef] = None,
    name: typing.Optional[str] = None,
    setter: typing.Optional[typing.Callable] = None,
    render_info: typing.Optional[dict] = None,
    hidden: bool = False,
    pure: bool = True,
    _op_def_class: type[op_def.OpDef] = op_def.OpDef,
    *,  # Marks the rest of the arguments as keyword-only.
    plugins: typing.Optional[dict[str, "GqlOpPlugin"]] = None,
    mutation: bool = False,
    # If True, the op will be weavified, ie it's resolver will be stored as a Weave
    # op graph. The compile node_ops pass will expand the node to the weavified
    # version, instead of executing the original python resolver body.
    weavify: bool = False,
) -> typing.Callable[[typing.Callable], op_def.OpDef]:
    """Decorator for declaring an op.

    Decorated functions must be typed, either with Python types or by declaring
    input_type, output_type as arguments to op (Python types preferred).
    """

    # For builtins, enforce that parameter and return types must be declared.
    # For user ops, allow missing types.
    allow_unknowns = not context_state._loading_built_ins.get()

    def wrap(f):
        weave_input_type = pyfunc_type_util.determine_input_type(
            f, input_type, allow_unknowns=allow_unknowns
        )
        weave_output_type = pyfunc_type_util.determine_output_type(
            f, output_type, allow_unknowns=allow_unknowns
        )

        fq_op_name = name
        if fq_op_name is None:
            op_prefix = "op"

            # This would be a much nicer: automatically use first type name.
            # But what about Unions? I'm not going to solve in this PR.
            # TODO: implement this.
            # if (
            #     isinstance(weave_input_type, op_args.OpNamedArgs)
            #     and weave_input_type.arg_types
            # ):
            #     arg_type0 = next(iter(weave_input_type.arg_types.values()))
            #     if arg_type0 != types.UnknownType():
            #         op_prefix = arg_type0.name

            fq_op_name = f"{op_prefix}-{f.__name__}"

        op = _op_def_class(
            fq_op_name,
            weave_input_type,
            weave_output_type,
            f,
            refine_output_type=refine_output_type,
            setter=setter,
            render_info=render_info,
            hidden=hidden,
            pure=pure,
            plugins=plugins,
            mutation=mutation,
            _decl_locals=inspect.currentframe().f_back.f_locals,
        )
        if weavify:
            from .weavify import op_to_weave_fn

            op.weave_fn = op_to_weave_fn(op)

        op_version = registry_mem.memory_registry.register_op(op)

        # After we register the op, create any derived ops

        # If op.location is set, then its a custom (non-builtin op). We don't
        # derive custom ops for now, as the derive code doesn't do the right thing.
        # The op name is the location/uri for custom ops, and the derive code doesn't
        # fix that up. So we end up double registering ops in WeaveJS which breaks
        # everything.
        # TODO: fix so we get mappability for custom (ecosystem) ops!
        if op.location is None:
            derive_op.derive_ops(op)

        return op_version

    return wrap



import typing
import weave

from .dataset import Dataset
from .model import Model
from .evaluate import Evaluate


@weave.op()
def evaluate(eval: Evaluate, dataset: Dataset, model: Model) -> typing.Any:
    outputs = dataset.rows.apply(lambda r: model.predict(r))  # type: ignore
    eval_result = eval.compute(dataset, outputs)
    summary = eval_result["summary"]

    eval_table_columns: dict[str, weave.WeaveList] = {
        "dataset_id": weave.WeaveList([r["id"] for r in dataset.rows]),
        # For now, put the whole dataset into the eval data. In the future we should put
        # refs instead.
        # TODO: use refs instead
        "dataset": dataset.rows,  # type: ignore
        "output": outputs,
        **eval_result["columns"],
    }
    return {
        "summary": summary,
        "eval_table": weave.WeaveList(eval_table_columns),
    }
